'use strict'

/**
 * @module
 */

import $ from 'jquery'

// ----------------------------------------
// Private
// ----------------------------------------

/**
 * @param {string} prop
 * @param {*} value
 * @private
 */
function isProtected (prop, value) {
  return console.warn(`ModuleLoader.${prop} is protected property! You cannot change it to "${value}" after create!`)
}

// ----------------------------------------
// Public
// ----------------------------------------

class ModuleLoader {
  /**
   * @param {Object} loaderData
   * @param {Promise} [loaderData.importPromise=null]
   * @param {string} [loaderData.initSelector='.js-init']
   * @param {string} [loaderData.initFunctionName='loaderInit']
   * @param {Object} [loaderData.list={}]
   * @param {Object} [loaderData.debug=false]
   * @param {string} [loaderData.loadingClass='_module-loading']
   * @param {string} [loaderData.loadedClass='_module-loaded']
   *
   * @prop {string} initSelector
   * @prop {string} initFunctionName
   * @prop {Object} list
   * @prop {Object} modules
   * @prop {boolean} debug
   * @prop {string} loadingClass
   * @prop {string} loadedClass
   */
  constructor (loaderData) {
    let {
      importPromise = null,
      initSelector = '.js-init',
      initFunctionName = 'loaderInit',
      list = {},
      loadingClass = '_module-loading',
      loadedClass = '_module-loaded',
      debug = false
    } = loaderData

    if (typeof importPromise !== 'function') {
      throw new Error('ModuleLoader ERROR! You must specify "importPromise" property, as Function with import()')
    }

    if (typeof list !== 'object' || Object.keys(list).length === 0) {
      throw new Error('ModuleLoader ERROR! You must specify "list" property, as Object with values')
    }

    for (let moduleName in list) {
      let selectors = list[moduleName]
      if (Array.isArray(selectors)) {
        list[moduleName] = selectors.join(',')
      }
    }

    const modules = {}
    for (let moduleName in list) {
      modules[moduleName] = () => importPromise(moduleName)
    }

    Object.defineProperty(this, 'initSelector', {
      get: () => initSelector,
      set: val => isProtected('initSelector', val)
    })

    Object.defineProperty(this, 'initFunctionName', {
      get: () => initFunctionName,
      set: val => isProtected('initFunctionName', val)
    })

    Object.defineProperty(this, 'list', {
      get: () => list,
      set: val => isProtected('list', val)
    })

    Object.defineProperty(this, 'modules', {
      get: () => modules,
      set: val => isProtected('modules', val)
    })

    Object.defineProperty(this, 'loadingClass', {
      get: () => loadingClass,
      set: val => isProtected('loadingClass', val)
    })

    Object.defineProperty(this, 'loadedClass', {
      get: () => loadedClass,
      set: val => isProtected('loadedClass', val)
    })

    this.debug = debug
  }

  /**
   * @param {string} type
   * @param {string} msg
   * @param {...*} [data]
   * @private
   */
  log (type, msg, ...data) {
    if (this.debug) {
      console[type](`ModuleLoader ${type}: ${msg}`, ...data)
    }
  }

  /**
   * @param {jQuery} [$context=$(document)]
   * @param {string} [selectors]
   * @returns {jQuery}
   * @public
   */
  getElements ($context = $(document), selectors) {
    let selector = this.initSelector
    if (selectors) {
      selector = selectors.split(',').map(selector => this.initSelector + selector).join(',')
    }
    this.log('info', `Get elements "${selector}" in`, $context)
    return $context.find(selector)
  }

  /**
   * @param {string} moduleName
   * @param {jQuery} [$context]
   * @public
   */
  runModule (moduleName, $context) {
    this.log('warn', 'Use .load() only method after .init()')
    console.log(moduleName, $context)
  }

  /**
   * @param {jQuery} $elements
   * @private
   */
  markAsLoadingElements ($elements) {
    $elements.not(this.loadedClass).addClass(this.loadingClass)
  }

  /**
   * @param {jQuery} $elements
   * @private
   */
  markAsLoadedElements ($elements) {
    $elements.removeClass(this.loadingClass).addClass(this.loadedClass)
    $elements.data('moduleLoader', this)
  }

  /**
   * @param {jQuery} [$context=$(document)]
   * @public
   */
  init ($context = $(document)) {
    const $elements = this.getElements($context)
    if (!$elements.length) {
      this.log('info', `No init elements with selector "${this.initSelector}"`)
      return false
    }

    this.runModule = function (moduleName, $context) {
      const selectors = this.list[moduleName]
      const $moduleElements = $context ? this.getElements($context, selectors) : $elements.filter(selectors)
      this.markAsLoadingElements($moduleElements)

      if ($moduleElements.length) {
        this.log('info', `Load module "${moduleName}"`)
        return this.modules[moduleName]().then(module => {
          const isFunction = typeof module[this.initFunctionName] === 'function'
          if (isFunction) {
            this.log('info', `Init module "${moduleName}"`)
            this.markAsLoadedElements($moduleElements)
            module[this.initFunctionName]($moduleElements, this)
          } else {
            this.log('warn', `Module "${moduleName}" has no exported method "${this.initSelector}"`)
          }
        }).catch(err => {
          console.error(err)
        })
      }

      return Promise.resolve()
    }

    for (let moduleName in this.list) {
      this.runModule(moduleName)
    }
  }
}

// ----------------------------------------
// Exports
// ----------------------------------------

export default ModuleLoader
